# -*- coding: utf-8 -*-

#  goma
#  ----
#  Generic object mapping algorithm.
#
#  Author:  pbrisk <pbrisk_at_github@icloud.com>
#  Copyright: 2016, 2017 Deutsche Postbank AG
#  Website: https://github.com/pbrisk/goma
#  License: APACHE Version 2 License (see LICENSE file)


from basematch import BaseMatch
from exactmatch import ExactMatch


class PriorityMatch(BaseMatch):
    """ Priority Match class """

    def __init__(self):
        self.exact_match = ExactMatch()

    def match(self, match_details, mapping_list, start_col=0):
        """ matching based on prioritizing entries in the right column of the
            mapping list

        Parameters:
            match_details (list): holds the information based on which the mapping
                                  should be conducted, a row entry is structured as
                                  [Detail ,Value]
            mapping_list (list): hold the mapping information, the first row describes
                                 the properties on which mapping should be conducted and
                                 a column named target
            start_col (int): starting column for the matching algorithms of the mapping_list

        Return:
            string: target value which has been matched


        The priority match uses a given mapping_list, e.g

        +-------------------------+------------+----------+
        | Property1  | Property2  | Property3  | Target   |
        +============+============+============+==========+
        | Value1_1   |            | Value3_1   | Target1  |
        +------------+------------+------------+----------+
        | Value1_1   |            | Value3_2   | Target2  |
        +------------+------------+------------+----------+
        |            | Value2_1   | Value3_3   | Target3  |
        +------------+------------+------------+----------+
        |            | Value2_1   | Value3_4   | Target4  |
        +------------+------------+------------+----------+

        Given the above tables, the matching searches column by column,
        starting on the left hand side (highest priority columns), for
        the respective mapping list.

        For a given list of match_details, e.g.

        +------------+-----------+
        | Property1  | Value1_2  |
        +------------+-----------+
        | Property2  | Value2_1  |
        +------------+-----------+
        | Property3  | Value3_3  |
        +------------+-----------+

        the match returns in a first step a 2x4 matrix

        +------------+------------+------------+----------+
        |            | Value2_1   | Value3_3   | Target3  |
        +------------+------------+------------+----------+
        |            | Value2_1   | Value3_4   | Target4  |
        +------------+------------+------------+----------+

        As a second step, the empty columns are removed and the exact match
        algorithm is applied to the remaining properties to find the target,
        which for the given example yields Target3.

        If for column no matching is found it continues recursively at the
        next column and searches for the respective property, i.e Property3.
        """

        match = False
        rows = []

        mapping_list_cols = list(mapping_list[0])
        target_col = mapping_list_cols.index('Target')

        # todo add some description on what is going on here
        for j, attr in enumerate(mapping_list_cols[start_col:-1]):
            for i in range(1, len(mapping_list)):
                detail_value = None
                for c in range(0, len(match_details)):
                    if match_details[c][0] == attr:
                        detail_value = match_details[c][1]
                        break

                if detail_value != '':
                    if self._has_relation_operator(mapping_list[i][j]):
                        if self._apply_relation_match(mapping_list[i][j], detail_value):
                            rows.append(i)
                            match = True
                        else:
                            match = False
                    elif type(detail_value)(mapping_list[i][j]) == detail_value:
                        rows.append(i)
                        match = True
                    else:
                        match = False
                else:
                    match = False

        new_mapping_list = list()
        new_mapping_list.append(mapping_list_cols[start_col:])
        new_mapping_list.extend([mapping_list[i][start_col:] for i in rows])

        # todo add some description on what is going on here
        if len(new_mapping_list[:-1]) == 1:
            match_obj = new_mapping_list[len(new_mapping_list) - 1][target_col]
        else:
            empty_cols = [[i for i, str in enumerate(j[:]) if str == ''] for j in new_mapping_list[1:]]
            unique_empty_cols = set([item for sublist in empty_cols for item in sublist])

            n = 0
            # todo add some description on what is going on here
            for i, j in enumerate(unique_empty_cols):
                if i == 0:
                    n = 0
                    [[new_mapping_list[r].pop(j)] for r in range(0, len(new_mapping_list))]
                else:
                    n += 1
                    [[new_mapping_list[r].pop(j - n)] for r in range(0, len(new_mapping_list))]

            match_obj = self.exact_match.match(match_details, new_mapping_list)

            if not match_obj:
                new_mapping_list_cols = [map_col for map_col in new_mapping_list[0]]
                start_col = [i for i, name in enumerate(mapping_list_cols) if name == new_mapping_list_cols[0]][0] + 1

                if start_col < len(mapping_list_cols[:-1]):
                    match_obj = self.match(match_details, mapping_list, start_col)

        return match_obj



# -*- coding: utf-8 -*-

#  goma
#  ----
#  Generic object mapping algorithm.
#
#  Author:  pbrisk <pbrisk_at_github@icloud.com>
#  Copyright: 2016, 2017 Deutsche Postbank AG
#  Website: https://github.com/pbrisk/goma
#  License: APACHE Version 2 License (see LICENSE file)


class BaseMatch(object):
    """ Base match class """

    def match(self, *args):
        """ return match results """
        pass  # functions are implemented in sub classes

    def _has_relation_operator(self, condition_str):
        return str(condition_str)[:1] in ['<', '>', '!', '=', '(', '[']

    def _apply_relation_match(self, match_value, detail_value):
        """ return relation match results """

        match_value = match_value.replace(' ', '')

        # if interval
        if match_value[:1] in ['[', '(']:
            rel_vals = [type(detail_value)(val) for val in match_value[1:-1].split(',')]
            rel_ops = [self._repl_relation_operator(op) for op in [match_value[0], match_value[-1]]]
            # fixme NO eval usage !!!
            match_res = eval(
                'detail_value ' + rel_ops[0] + ' rel_vals[0] and detail_value ' + rel_ops[1] + ' rel_vals[1]')
        # else
        else:
            if self._repl_relation_operator(match_value[:2]):
                rel_ops = match_value[:2]
                rel_val = type(detail_value)(match_value[2:])
            else:
                rel_ops = match_value[:1]
                rel_val = type(detail_value)(match_value[1:])

            match_res = eval('detail_value ' + rel_ops + ' rel_val')

        return match_res

    def _repl_relation_operator(self, relation_symbol):
        if relation_symbol == '[': return '>='
        if relation_symbol == '(': return '>'
        if relation_symbol == ']': return '<='
        if relation_symbol == ')': return '<'
        if relation_symbol in ['<=', '=<']: return '<='
        if relation_symbol in ['>=', '=>']: return '>='
        if relation_symbol == '==': return '=='
        if relation_symbol == '!=': return '!='
        return relation_symbol



# -*- coding: utf-8 -*-

#  goma
#  ----
#  Generic object mapping algorithm.
#
#  Author:  pbrisk <pbrisk_at_github@icloud.com>
#  Copyright: 2016, 2017 Deutsche Postbank AG
#  Website: https://github.com/pbrisk/goma
#  License: APACHE Version 2 License (see LICENSE file)


from basematch import BaseMatch


class ExactMatch(BaseMatch):
    def match(self, match_details, mapping_list):
        """ matching based on exact entries of all columns of the
            mapping list

        Parameters:
            match_details (list): holds the information based on which the mapping
                                  should be conducted, a row entry is structured as
                                  ['Detail', 'Value']
            mapping_list (list): holds the mapping information, the first row describes
                                 the properties on which mapping should be conducted and
                                 a column named target
            start_col (int): starting column for the matching algorithms of the mapping_list

        The exact match uses a given mapping_list, e.g

        +-------------------------+------------+----------+
        | Property1  | Property2  | Property3  | Target   |
        +============+============+============+==========+
        | Value1_1   |  Value2_1  | Value3_1   | Target1  |
        +------------+------------+------------+----------+
        | Value1_2   |  Value2_2  | Value3_2   | Target2  |
        +------------+------------+------------+----------+

        Given the above tables, the matching searches row by row,
        if all criteria in the matching list of a given list of
        details (match_details) are met.

        For a given list of match_details, e.g.

        +------------+-----------+
        | Property1  | Value1_2  |
        +------------+-----------+
        | Property2  | Value2_2  |
        +------------+-----------+
        | Property3  | Value3_2  |
        +------------+-----------+

        the match returns in a Target 2. If one matching criteria is not
        met, the match returns None.
        """

        match_obj = None

        mapping_list_cols = [map_col for map_col in mapping_list[0]]
        target_col = [i for i, name in enumerate(mapping_list[0]) if name == 'Target'][0]

        for i in range(1, len(mapping_list)):
            matches = 0
            for j, attr in enumerate(mapping_list_cols[:-1]):
                detail_value = None
                for c in range(0, len(match_details)):
                    if match_details[c][0] == attr:
                        detail_value = match_details[c][1]
                        break

                if detail_value != '':
                    if self._has_relation_operator(mapping_list[i][j]):
                        if self._apply_relation_match(mapping_list[i][j], detail_value):
                            matches += 1
                        else:
                            break
                    elif not detail_value:
                        break
                    elif type(detail_value)(mapping_list[i][j]) == detail_value:
                        matches += 1
                    else:
                        break
                else:
                    break

            if matches == len(mapping_list_cols[:-1]):
                match_obj = mapping_list[i][target_col]
                break

        return match_obj



# -*- coding: utf-8 -*-

#  goma
#  ----
#  Generic object mapping algorithm.
#
#  Author:  pbrisk <pbrisk_at_github@icloud.com>
#  Copyright: 2016, 2017 Deutsche Postbank AG
#  Website: https://github.com/pbrisk/goma
#  License: APACHE Version 2 License (see LICENSE file)


import codecs
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


setup(
    name='goma',
    description='Python library for genaralized object mapping algorithms.',
    version='0.1',
    author='Deutsche Postbank AG [pbrisk]',
    author_email='pbrisk_at_github@icloud.com',
    url='https://github.com/pbrisk/goma',
    bugtrack_url='https://github.com/pbrisk/goma/issues',
    license='Apache License 2.0',
    packages=['goma'],
    long_description=codecs.open('README.rst', encoding='utf-8').read(),
    platforms='any',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Financial and Insurance Industry',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Education',
        'Topic :: Office/Business',
        'Topic :: Office/Business :: Financial',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Utilities',
        'Topic :: Office/Business :: Scheduling',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Localization',
    ],
)



# -*- coding: utf-8 -*-

#  goma
#  ----
#  General object mapping algorithm.
#
#  Author:  pbrisk <pbrisk_at_github@icloud.com>
#  Copyright: 2016, 2017 Deutsche Postbank AG
#  Website: https://github.com/pbrisk/goma
#  License: APACHE Version 2 License (see LICENSE file)


import os
import unittest
from goma import ExactMatch, PriorityMatch

from datetime import datetime, date, timedelta

class ExactMatchUnitTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_match(self):
        exact_match = ExactMatch()

        detail_list = [["Property1","Value1_2"],["Property2","Value2_2"],["Property3","Value3_2"]]
        mapping_list = [["Property1","Property2","Property3","Target"], \
                        ["Value1_1","Value2_1","Value3_1","Target1"], \
                        ["Value1_2","Value2_2","Value3_2","Target2"]]

        result = exact_match.match(detail_list, mapping_list)
        self.assertEqual("Target2",result, "Mapping algorithm not set up properly.")

class PriorityMatchUnitTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_match(self):
        priority_match = PriorityMatch()

        detail_list = [["Property1","Value1_2"],["Property2","Value2_1"],["Property3","Value3_3"]]
        mapping_list = [["Property1","Property2","Property3","Target"], \
                        ["Value1_1","","Value3_1","Target1"], \
                        ["Value1_1","","Value3_2","Target2"], \
                        ["","Value2_1","Value3_3","Target3"], \
                        ["","Value2_1","Value3_4","Target4"]]

        result = priority_match.match(detail_list, mapping_list)
        self.assertEqual("Target3",result, "Mapping algorithm not set up properly.")

if __name__ == "__main__":

    import sys
    start_time = datetime.now()

    print('')
    print('======================================================================')
    print('')
    print('run %s' % __file__)
    print('in %s' % os.getcwd())
    print('started  at %s' % str(start_time))
    print('')
    print('----------------------------------------------------------------------')
    print('')

    suite = unittest.TestLoader().loadTestsFromModule(__import__("__main__"))
    testrunner = unittest.TextTestRunner(stream=sys.stdout, descriptions=2, verbosity=2)
    testrunner.run(suite)

    print('')
    print('======================================================================')
    print('')
    print('ran %s' % __file__)
    print('in %s' % os.getcwd())
    print('started  at %s' % str(start_time))
    print('finished at %s' % str(datetime.now()))
    print('')
    print('----------------------------------------------------------------------')
    print('')
